package de.dfki.madm.mlwizard;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.rapidminer.Process;
import com.rapidminer.example.Attribute;
import com.rapidminer.example.AttributeWeights;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.ExampleSetFactory;
import com.rapidminer.operator.AbstractIOObject;
import com.rapidminer.operator.Annotations;
import com.rapidminer.operator.IOContainer;
import com.rapidminer.operator.Model;
import com.rapidminer.operator.OperatorCreationException;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.io.XrffExampleSource;
import com.rapidminer.operator.learner.AbstractLearner;
import com.rapidminer.operator.meta.ParameterSet;
import com.rapidminer.operator.meta.ParameterValue;
import com.rapidminer.operator.performance.PerformanceVector;
import com.rapidminer.operator.preprocessing.filter.NominalNumbers2Numerical;
import com.rapidminer.operator.preprocessing.normalization.Normalization;
import com.rapidminer.tools.OperatorService;
import com.rapidminer.tools.RandomGenerator;
import com.rapidminer.tools.XMLException;
import com.rapidminer.tools.container.Pair;

import de.dfki.madm.mlwizard.functionality.Evaluator;

public class Classifier extends AbstractIOObject {
    
    Logger logger = Logger.getLogger(Classifier.class.getName());
    
    private static final long serialVersionUID = 3538346855933324659L;
    
    private final String name;
    private boolean complex;
    
    private final Class<? extends AbstractLearner> learner;
    private final String preProcess, learnProcess;

    private final List<String[]> numericalParams, nominalParams;
    
    private ExampleSet parameters, accuracies;
    
    private Model regressionModel;
    private PerformanceVector rmse;
    private Model preprocessingModel;
    private AttributeWeights metaWeightsRegression, metaWeightsParameters;
    private double paramDist;

    public Classifier(String name, Class<? extends AbstractLearner> learner, List<String[]> bounds, List<String[]> nominalParams) {
        this.complex = false;
        this.name = name;
        this.learner = learner;
        this.numericalParams = new LinkedList<String[]>(bounds);
        this.nominalParams = new LinkedList<String[]>(nominalParams);
        this.paramDist = Double.MAX_VALUE;
        this.preProcess = null;
        this.learnProcess = null;
        this.accuracies = null;
        this.parameters = null;
    }
    
    public Classifier(String name, String preProcess, String learnProcess, List<String[]> bounds, List<String[]> nominalParams) {
        this.complex = true;
        this.name = name;
        this.learner = null;
        this.numericalParams = new LinkedList<String[]>(bounds);
        this.nominalParams = new LinkedList<String[]>(nominalParams);
        this.paramDist = Double.MAX_VALUE;
        this.preProcess = preProcess;
        this.learnProcess = learnProcess;
        this.accuracies = null;
        this.parameters = null;
    }
    
    public Classifier(String name, Class<? extends AbstractLearner> learner, List<String[]> bounds, List<String[]> nominalParams, Classifier other) {
        this(name, learner, bounds, nominalParams);
        if(other != null) {
            this.complex = other.complex;
            this.parameters = other.parameters;
            this.accuracies = other.accuracies;
            this.regressionModel = other.regressionModel;
            this.rmse = other.rmse;
            this.preprocessingModel = other.preprocessingModel;
            this.metaWeightsParameters  = other.metaWeightsParameters;
            this.metaWeightsRegression = other.metaWeightsRegression;
        }
    }
    
    public boolean isComplex() {
        return complex;
    }
    
    public void addAccuracy(String datasetName, Double accuracy) throws OperatorCreationException, OperatorException {
        ExampleSet newData = ExampleSetFactory.createExampleSet(new String[][]{{datasetName}}, new Double[]{accuracy});
        newData.getAttributes().setId(newData.getAttributes().get("att1"));
        newData.getAttributes().getId().setName("dataset");
        newData.getAttributes().getLabel().setName("accuracy");
        accuracies = Util.mergeExampleSets(accuracies, newData);
    }
    
    public void addParameters(String datasetName, ParameterSet parameters) throws OperatorCreationException, OperatorException {
        Vector<String> data = new Vector<String>();
        data.add(datasetName);
        for(Iterator<ParameterValue> it=parameters.getParameterValues(); it.hasNext();) {
            data.add(it.next().getParameterValue());
        }
        ExampleSet newData = ExampleSetFactory.createExampleSet(new Object[][]{data.toArray(new String[0])});
        newData.getAttributes().get("att1").setName("dataset");
        int i=2;
        for(Iterator<ParameterValue> it=parameters.getParameterValues(); it.hasNext();) {
            newData.getAttributes().get("att"+(i++)).setName(this.name + "." + it.next().getParameterKey());
        }
        newData.getAttributes().setId(newData.getAttributes().get("dataset"));
        NominalNumbers2Numerical parser = OperatorService.createOperator(NominalNumbers2Numerical.class);
        this.parameters = Util.mergeExampleSets(this.parameters, parser.apply(newData));
    }
    
    public String getName() {
        return name;
    }
    
    public Model getPreprocessingModel() {
        return preprocessingModel;
    }
    
    public Model getRegressionModel() {
        return regressionModel;
    }
    
    public ExampleSet getParameters() {
        return parameters;
    }

    public void setParameters(ExampleSet parameters) {
        this.parameters = parameters;
    }
    
    public void setParameters(String xrffFile) throws OperatorCreationException, OperatorException {
        XrffExampleSource reader = OperatorService.createOperator(XrffExampleSource.class);
        reader.setParameter(XrffExampleSource.PARAMETER_DATA_FILE, xrffFile);
        parameters = reader.read();
        parameters.getAttributes().setId(parameters.getAttributes().get("dataset"));
    }

    public ExampleSet getAccuracies() {
        return accuracies;
    }

    public void setAccuracies(ExampleSet accuracies) {
        this.accuracies = accuracies;
    }
    
    public void setAccuracies(String xrffFile) throws OperatorCreationException, OperatorException {
        XrffExampleSource reader = OperatorService.createOperator(XrffExampleSource.class);
        reader.setParameter(XrffExampleSource.PARAMETER_DATA_FILE, xrffFile);
        accuracies = reader.read();
        accuracies.getAttributes().setId(accuracies.getAttributes().get("dataset"));
    }
    
    public AttributeWeights getMetaWeightsParameters() {
        return metaWeightsParameters;
    }

    private void setMetaWeightsParameters(long mask, double paramDist, ExampleSet metaFeatures) {
        if(this.metaWeightsParameters == null || paramDist < this.paramDist) {
            logger.log(Level.INFO, String.format("New min param dist for %s: %f %d", getName(), Math.sqrt(paramDist), mask));
            // -- construct attribute weights out of best mask --
            metaWeightsParameters = new AttributeWeights(metaFeatures);
            long i = 0;
            for(Attribute attribute:metaFeatures.getAttributes()) {
                if((mask & (1l << i)) != 0) {
                    metaWeightsParameters.setWeight(attribute.getName(), 1.0);
                } else {
                    metaWeightsParameters.setWeight(attribute.getName(), 0.0);
                }
                i++;
            }
            this.paramDist = paramDist;
        }
    }

    public AttributeWeights getMetaWeightsRegression() {
        return metaWeightsRegression;
    }

    public double getRMSE() {
        return rmse.getMainCriterion().getAverage();
    }
    
    public List<String[]> getNumericalParams() {
        return numericalParams;
    }
    
    public List<String[]> getNominalParmas() {
        return nominalParams;
    }
    
    public Class<? extends AbstractLearner> getLearner() {
        return learner;
    }
    
    public AbstractLearner getLeanerInstance() throws OperatorCreationException {
        return OperatorService.createOperator(learner);
    }
    
    public String getPreProcess() {
        return preProcess;
    }

    public String getLearnProcess() {
        return learnProcess;
    }
    
    public Example getParameters(String dataset) {
        for(Example example:parameters) {
            if(example.getNominalValue(example.getAttributes().getId()).equals(dataset)) {
                return example;
            }
        }
        return null;
    }
    
    public int getOptimizationCase() {
        int c = 1;
        if(!nominalParams.isEmpty()) {
            c = 2;
        }
        if(!numericalParams.isEmpty()) {
            if(c==2) {
                c = 4;
            } else {
                c = 3;
            }
        }
        return c;
    }
    
    public void trainRegressionModel(ExampleSet metaFeatures) throws IOException, XMLException, OperatorException {
        Process process = new Process(new File("resources/com/rapidminer/resources/data/learn_regression_model.xml"));
        IOContainer results = process.run(new IOContainer(accuracies, metaFeatures));
        this.preprocessingModel = results.get(Model.class, 0);
        this.regressionModel = results.get(Model.class, 1);
        this.metaWeightsRegression = results.get(AttributeWeights.class);
        this.rmse = results.get(PerformanceVector.class);
    }
    
    public void determineOptimalMetaFeaturesForParameterPrediction(ExampleSet metaFeatures) throws IOException, OperatorCreationException, OperatorException {
        final int N = 1000;
        // -- normalize parameters --
        Normalization normalization = OperatorService.createOperator(Normalization.class);
        normalization.setParameter(Normalization.PARAMETER_NORMALIZATION_METHOD, String.valueOf(Normalization.METHOD_RANGE_TRANSFORMATION));
        ExampleSet normalizedParameters = normalization.doWork(this.parameters);

        Attribute idAttributeParameters = normalizedParameters.getAttributes().getId();
        
        // -- copy meta-features and parameters to map for faster access --
        HashMap<String, Pair<Example, Example> > metaFeaturesMap = new HashMap<String, Pair<Example, Example>>();
        Vector<String> datasetNames = new Vector<String>();
        for(Example p:normalizedParameters) {
            String datasetName = p.getNominalValue(idAttributeParameters);
            Example m = Util.getExampleFromId(metaFeatures, datasetName);
            // -- only if the meta-features of this dataset exist --
            if(m != null) {
                metaFeaturesMap.put(datasetName, new Pair<Example, Example>(p, m));
                datasetNames.add(datasetName);
            }
        }
        
        // -- loop over each attribute subset --
        long highestMask = (1l << (long)metaFeatures.getAttributes().size());
//        for(long mask=1l; mask<highestMask; mask++) {
        RandomGenerator r = new RandomGenerator(System.currentTimeMillis());
        for(int k=0; k<N; k++) {
            if(k%1000 == 0) {
                logger.log(Level.INFO, String.format("%.3f", (k*100.0/N)));
            }
            long mask = r.nextLongInRange(0, highestMask);
            // -- leave-one-out cross-validation --
            double paramDist = 0.0;
            for(int i=0; i<datasetNames.size()-1; i++) {
                String queryDatasetName = datasetNames.get(i);
                Example queryParameters = metaFeaturesMap.get(queryDatasetName).getFirst();
                Example queryMetaFeatures = metaFeaturesMap.get(queryDatasetName).getSecond();
                if(queryMetaFeatures == null) {
                    continue;
                }
                // -- find nearest neighbor --
                double minFeatDist = Double.MAX_VALUE;
                Example nearestParameters = null;
                for(int j=i+1; j<datasetNames.size(); j++) {
                    String otherDatasetName = datasetNames.get(j);
                    Example otherParameters = metaFeaturesMap.get(otherDatasetName).getFirst();
                    Example otherMetaFeatures = metaFeaturesMap.get(otherDatasetName).getSecond();
                    if(otherMetaFeatures == null) {
                        continue;
                    }
                    double featDist = Util.dist(queryMetaFeatures, otherMetaFeatures, mask);
                    if(nearestParameters == null || featDist < minFeatDist) {
                        minFeatDist = featDist;
                        nearestParameters = otherParameters;
                    }
                }
                // -- determine distance in parameter space --
                paramDist += Util.dist(queryParameters, nearestParameters);
            }
            setMetaWeightsParameters(mask, paramDist, metaFeatures);
        }
        logger.log(Level.INFO, getName() + " finished");
    }
    
    public void evaluate(KnowledgeBase knowledgeBase) {
        Vector<Classifier> classifiers = new Vector<Classifier>();
        classifiers.add(this);
        Evaluator evaluator = new Evaluator();
        for(String datasetName:knowledgeBase.getStoredDatasets()) {
            // -- if accuracy result is missing for  that dataset --
            if(accuracies == null || Util.getExampleFromId(accuracies, datasetName) == null) {
                try {
                    logger.log(Level.INFO, "evaluate '" + name + "' on '" + datasetName + "'");
                    ExampleSet dataset = KnowledgeBase.readDataSet(datasetName);
                    Example metaFeature = Util.getExampleFromId(knowledgeBase.getMetaFeatures(), datasetName);
                    Map<Classifier, ParameterSet> results = evaluator.evaluate(classifiers, metaFeature, knowledgeBase, dataset, 1, null);
                    evaluator.awaitTermination();
                    ParameterSet parameters = results.get(this);
                    addAccuracy(datasetName, parameters.getPerformance().getMainCriterion().getFitness());
                    addParameters(datasetName, parameters);
                } catch (Exception e) {
                    e.printStackTrace();
                    System.exit(-1);
                }
            } else {
                logger.log(Level.INFO, "skip evaluating '" + name + "' on '" + datasetName + "'");
            }
        }
    }

    @Override
    public Annotations getAnnotations() {
        return null;
    }
}
