package de.dfki.madm.mlwizard;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Vector;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.table.AttributeFactory;
import com.rapidminer.operator.AbstractIOObject;
import com.rapidminer.operator.Annotations;
import com.rapidminer.operator.OperatorCreationException;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.meta.ParameterSet;
import com.rapidminer.operator.preprocessing.filter.MissingValueReplenishment;
import com.rapidminer.tools.Ontology;
import com.rapidminer.tools.OperatorService;
import com.rapidminer.tools.XMLException;
import com.rapidminer.tools.container.Pair;

import de.dfki.madm.mlwizard.functionality.Evaluator;
import de.dfki.madm.mlwizard.metafeatures.MetaFeaturesOperator;

/**
 * Central place for all knowledge regarding meta-learning including meta-features and classifiers.
 * Administrative functionality like adding new datasets is shifted to the Administrator class for clarity reasons. 
 */
public class KnowledgeBase extends AbstractIOObject {

    private static final long serialVersionUID = 2293093005413914177L;
    
    /** the location of the knowledge base access as an resource */
    private static final String RESOURCE_NAME = "knowledgebase";
    
    /** the meta-features of all datasets */
    private ExampleSet metaFeatures;
    /** the classifiers accessible via their names */
    private HashMap<String, Classifier> classifiers;
    private Vector<String> storedDatasets;

    /** thread pool for parallel execution of learning and feature selection */
    private static ExecutorService threads = Executors.newFixedThreadPool(6/*Runtime.getRuntime().availableProcessors()*/);
    
    public Collection<String> getStoredDatasets() {
        return Collections.unmodifiableCollection(storedDatasets);
    }

    /**
     * creates an empty knowledge base
     */
    public KnowledgeBase() {
        classifiers = new HashMap<String, Classifier>();
        storedDatasets = new Vector<String>();
        metaFeatures = null;
    }
    
    /**
     * @param name name of the desired classifier
     * @return the classifier with the given name or null if no such classifier exists
     */
    public Classifier getClassifier(String name) {
        return classifiers.get(name);
    }
    
    /**
     * @return collection of all classifiers
     */
    public final Collection<Classifier> getClassifiers() {
        return classifiers.values();
    }
    
    /**
     * adds a classifier to this knowledge base
     * @param classifier the classifier to be added
     * @throws OperatorException 
     * @throws XMLException 
     * @throws IOException 
     * @throws OperatorCreationException 
     */
    public void addClassifier(Classifier classifier, boolean evaluate) throws IOException, XMLException, OperatorException, OperatorCreationException {
        classifiers.put(classifier.getName(), classifier);
        if(evaluate) {
            threads.execute(new LearningThread(classifier, this, evaluate));
        }
    }
    
    /**
     * @return all meta-features currently stored within this knowledge base
     */
    public ExampleSet getMetaFeatures() {
        return metaFeatures;
    }
    
    /**
     * replaces the meta-features
     * @param metaFeatures the new meta-features
     */
    public void setMetaFeatures(ExampleSet metaFeatures) {
        this.metaFeatures = metaFeatures;
    }
    
    /**
     * appends new meta-features to to currently existing ones
     * @param metaFeatures the meta-features to be added (one or more datasets)
     * @throws OperatorCreationException
     * @throws OperatorException
     */
    private void addMetaFeatures(ExampleSet metaFeatures) throws OperatorCreationException, OperatorException {
        if(this.metaFeatures == null) {
            this.metaFeatures = metaFeatures;
        } else {
            this.metaFeatures = Util.mergeExampleSets(this.metaFeatures, metaFeatures);
        }
    }
    
    /**
     * adds a new dataset to the knowledge base by computing its meta-features, evaluating all classifiers,
     * and updating the accuracy prediction models as well as the set of selected meta-features 
     * @param dataset the dataset to be added
     * @param name the name of the dataset used as id
     * @param knowledgeBase the knowledge base the dataset should be added to
     * @throws OperatorCreationException
     * @throws OperatorException
     * @throws IOException 
     * @throws FileNotFoundException 
     * @throws XMLException 
     * @throws InterruptedException 
     */
    public void addDataset(ExampleSet dataset, String name, boolean store) throws OperatorCreationException, OperatorException, FileNotFoundException, IOException, XMLException, InterruptedException {
        if(store) {
            storedDatasets.add(name);
            Util.write(Util.DATASETS_FOLDER + name, dataset);
        }
        MissingValueReplenishment missing = OperatorService.createOperator(MissingValueReplenishment.class);
        MetaFeaturesOperator mf = OperatorService.createOperator(MetaFeaturesOperator.class);
        dataset = missing.doWork(dataset);
        ExampleSet metaFeatures = mf.apply(dataset);
        Attribute datasetAttribute = AttributeFactory.createAttribute("dataset", Ontology.POLYNOMINAL);
        metaFeatures.getExampleTable().addAttribute(datasetAttribute);
        metaFeatures.getAttributes().addRegular(datasetAttribute);
        metaFeatures.getAttributes().setId(datasetAttribute);
        metaFeatures.getExample(0).setValue(datasetAttribute, name);
        addMetaFeatures(metaFeatures);
        
        Evaluator evaluator = new Evaluator();
        Map<Classifier, ParameterSet> results = evaluator.evaluate(classifiers.values(), metaFeatures.getExample(0), this, dataset, -1, null);
        evaluator.awaitTermination();
        
        ExecutorService threads = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());
        for(Entry<Classifier, ParameterSet> result:results.entrySet()) {
            Classifier classifier = result.getKey();
            classifier.addAccuracy(name, result.getValue().getPerformance().getMainCriterion().getAverage());
            classifier.addParameters(name, result.getValue());
            // -- update regression model and feature selection but do not evaluate the classifier again --
            threads.execute(new LearningThread(classifier, this, false));
        }
        threads.shutdown();
        threads.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
    }

    /**
     * reads the knowledge from the resource (used within the extension)
     * @return the knowledge base read 
     * @throws IOException
     */
    public static KnowledgeBase read() throws IOException {
        return Util.read(RESOURCE_NAME, KnowledgeBase.class);
    }
    
    
    /**
     * writes the knowledge base to the file (used within a stand-alone application)
     * @throws FileNotFoundException
     * @throws IOException
     */
    public void write() throws FileNotFoundException, IOException {
        Util.write(RESOURCE_NAME, this);
    }
    
    /**
     * reads a dataset either from resource or from file
     * @param name the name of the dataset
     * @return the dataset
     * @throws IOException
     */
    public static ExampleSet readDataSet(String name) throws IOException {
        return Util.read(Util.DATASETS_FOLDER + name, ExampleSet.class);
    }
    
    /**
     * determines the start points for the GA for a given classifier and dataset
     * @param classifier the classifier
     * @param queryMetaFeatures the meta-features of the dataset
     * @param n the desired number of start points 
     * @return the n points as flat list with all dimensions of the first point, then all dimension of the second point, ...
     */
    public List<Double> getStartPoints(Classifier classifier, Example queryMetaFeatures, int n) {
        // -- calculate distances to known datasets according to the selected meta-features --
        List< Pair<String, Double> > dists = new LinkedList< Pair<String, Double> >();
        for(Example example:classifier.getParameters()) {
            String id = example.getNominalValue(example.getAttributes().getId());
            Example m = Util.getExampleFromId(metaFeatures, id);
            if(m != null) {
                double d = Util.dist(queryMetaFeatures, m, classifier.getMetaWeightsParameters());
                dists.add(new Pair<String, Double>(id, d));
            }
        }
        // -- sort distances to known datasets --
        Collections.sort(dists, new Comparator<Pair<String, Double>>() {
            @Override
            public int compare(Pair<String, Double> o1, Pair<String, Double> o2) {
                return Double.compare(o1.getSecond(), o2.getSecond());
            }
        });
        // -- copy the optimal parameter values of the nearest datasets to a list for the '(not)-random' generator --
        List<Double> startPoints = new LinkedList<Double>();
        int c = 0;
        for(Pair<String, Double> e:dists) {
            Example parameterValues = classifier.getParameters(e.getFirst());
            for(String[] parameterNames:classifier.getNumericalParams()) {
                double xx = parameterValues.getValue(parameterValues.getAttributes().get(parameterNames[0]));
                startPoints.add(xx);
                Logger.getAnonymousLogger().log(Level.INFO, parameterNames[0] + ": " + xx);
            }
            if(++c == n) {
                break;
            }
        }
        return startPoints;
    }

    public void awaitTermination(long timeout, TimeUnit unit) throws InterruptedException {
        threads.awaitTermination(timeout, unit);
    }
    
    public static class LearningThread implements Runnable {
        private Classifier classifier;
        private KnowledgeBase knowledgeBase;
        private boolean evaluate;
        public LearningThread(Classifier classifier, KnowledgeBase knowledgeBase, boolean evaluate) {
            this.classifier = classifier;
            this.knowledgeBase = knowledgeBase;
            this.evaluate = evaluate;
        }
        @Override
        public void run() {
            try {
                if(evaluate) {
                    classifier.evaluate(knowledgeBase);
                }
                classifier.trainRegressionModel(knowledgeBase.getMetaFeatures());
                classifier.determineOptimalMetaFeaturesForParameterPrediction(knowledgeBase.getMetaFeatures());
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }
    
    @Override
    public Annotations getAnnotations() {
        return null;
    }
}

