package de.dfki.madm.mlwizard;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.AttributeWeights;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.operator.AbstractIOObject;
import com.rapidminer.operator.IOObject;
import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorCreationException;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.io.RepositorySource;
import com.rapidminer.operator.io.RepositoryStorer;
import com.rapidminer.operator.meta.ParameterSet;
import com.rapidminer.operator.meta.ParameterValue;
import com.rapidminer.operator.performance.PerformanceVector;
import com.rapidminer.operator.preprocessing.join.ExampleSetMerge;
import com.rapidminer.tools.OperatorService;
import com.rapidminer.tools.XMLException;

/**
 * @brief class providing some basic utility methods
 */
public class Util {
    
    public static final String RESOURCE_PATH = "com/rapidminer/resources/data/";
    public static final String FILE_PATH = "resources/" + RESOURCE_PATH;
    public static final String DATASETS_FOLDER = "datasets/";
    

    /**
     * gets the example with the given id using a linear search
     * @param exampleSet the dataset
     * @param id the desired id
     * @return the example with the id or null if no such example exists
     */
    public static Example getExampleFromId(ExampleSet exampleSet, String id) {
        for(Example example:exampleSet) {
            if(example.getNominalValue(example.getAttributes().getId()).equals(id)) {
                return example;
            }
        }
        return null;
    }
    
    /**
     * computes the squared euclidean distance between a and b (omits the square root calculation for efficiency reasons)
     * @param a example a
     * @param b example b
     * @return (pseudo) distance between a and b 
     */
    public static double dist(Example a, Example b) {
        return dist(a, b, 9223372036854775807l);  // 9223372036854775807 => 0b111111111111111111111111111111111111111111111111111111111111111
    }
    
    /**
     * computes the squared euclidean distance between a and b (omits the square root calculation for efficiency reasons) using the given set of attributes only
     * @param a example a
     * @param b example b
     * @param weights only attributes with a weight >=1 will be used within the distance calculation
     * @return the distance between a and b
     */
    public static double dist(Example a, Example b, AttributeWeights weights) {
        if(a==null) System.out.println("a");
        if(b==null) System.out.println("b");
        if(weights==null) System.out.println("weights");
        double dist = 0.0;
        for(Attribute attribute:a.getAttributes()) {
            String n = attribute.getName();
            if(attribute.isNumerical() && weights.getWeight(n)>=1.0) {
                dist += Math.pow(a.getNumericalValue(attribute) - b.getNumericalValue(b.getAttributes().get(n)), 2.0);
            }
        }
//        return Math.sqrt(dist);
        // -- we do not need the square root since we just need the ordering --
        return dist;
    }
    
    /**
     * computes the squared euclidean distance between a and b (omits the square root calculation for efficiency reasons) using the given binary mask of attribute
     * @param a example a
     * @param b example b
     * @param mask only attributes with the given bit in the mask set are used within the distance calculation (first attribute corresponds to lowest significant bit) 
     * @return the distance between a and b
     */
    public static double dist(Example a, Example b, long mask) {
        if(a==null) System.out.println("a");
        if(b==null) System.out.println("b");
        double dist = 0.0;
        long i = 0;
        for(Attribute attribute:a.getAttributes()) {
            if((mask & (1l << i)) != 0l && attribute.isNumerical()) {
                double diff = a.getNumericalValue(attribute) - b.getNumericalValue(attribute);
                dist += diff * diff;
            }
            i++;
        }
//      return Math.sqrt(dist);
      // -- we do not need the square root since we just need the ordering --
      return dist;
    }
    
    /**
     * combine two parameter sets
     * @param p1 parameter set 1
     * @param p2 parameter set 1
     * @param pv the performance of the new parameter set
     * @return parameter set containing all parameters from p1 and p2
     */
    public static ParameterSet combineParameterSets(ParameterSet p1, ParameterSet p2, PerformanceVector pv) {
        if(p2 != null) {
            Vector<String> operators = new Vector<String>();
            Vector<String> keys = new Vector<String>();
            Vector<String> values = new Vector<String>();
            for(Iterator<ParameterValue> it=p1.getParameterValues(); it.hasNext(); ) {
                ParameterValue p = it.next();
                operators.add(p.getOperator());
                keys.add(p.getParameterKey());
                values.add(p.getParameterValue());
            }
            for(Iterator<ParameterValue> it=p2.getParameterValues(); it.hasNext(); ) {
                ParameterValue p = it.next();
                operators.add(p.getOperator());
                keys.add(p.getParameterKey());
                values.add(p.getParameterValue());
            }
            return new ParameterSet(operators.toArray(new Operator[0]), keys.toArray(new String[0]), values.toArray(new String[0]), pv);
        } else {
            return p1;
        }
    }
    
    /**
     * reads an object from a resource or a file. First, reading from the resource is tried, if it fails, reading from the file is tried.
     * @param name
     * @param clazz
     * @return
     * @throws IOException 
     */
    public static <T extends IOObject> T read(final String name, Class<T> clazz) throws IOException {
        try {
            return readFromResource(RESOURCE_PATH + name, clazz);
        } catch (Exception e) {
            e.printStackTrace();
            return readFromFile(FILE_PATH + name, clazz);
        }
    }
    

    /**
     * reads an IOObject from a resource
     * @param resourcePath the resource location
     * @param clazz the desired class of the IOObject
     * @return the object read
     * @throws IOException
     */
    public static <T extends IOObject> T readFromResource(final String resourcePath, Class<T> clazz) throws IOException {
        return clazz.cast(AbstractIOObject.read(new AbstractIOObject.InputStreamProvider() {
            @Override
            public InputStream getInputStream() throws IOException {
                return Util.class.getClassLoader().getResourceAsStream(resourcePath);
            }
        }));
    }
    
    /**
     * reads an IOObject from a file
     * @param name the resource location
     * @param clazz the desired class of the IOObject
     * @return the object read
     * @throws IOException
     */
    public static <T extends IOObject> T readFromFile(final String filePath, Class<T> clazz) throws IOException {
        return clazz.cast(AbstractIOObject.read(new File(filePath)));
    }
    

    public static void write(String name, IOObject object) throws FileNotFoundException, IOException {
        object.write(new FileOutputStream(new File(FILE_PATH + name)));
    }

    
    public static InputStream getStream(final String name) throws IOException, XMLException {
        InputStream stream = null;
        try {
            stream = getStreamFromResource(RESOURCE_PATH + name);
        } catch (Exception e) {
        }
        if(stream == null) {
            stream = getStreamFromFile(FILE_PATH + name);
        }
        return stream;
    }
    
    public static InputStream getStreamFromResource(final String resourcePath) throws IOException, XMLException {
        return Util.class.getClassLoader().getResourceAsStream(resourcePath);
    }
    
    public static InputStream getStreamFromFile(final String filePath) throws IOException, XMLException {
        return new FileInputStream(filePath);
    }
    
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    
    public static void writeToRepository(IOObject object, String location) throws OperatorCreationException, OperatorException {
        RepositoryStorer storer = OperatorService.createOperator(RepositoryStorer.class);
        storer.setParameter(RepositoryStorer.PARAMETER_REPOSITORY_ENTRY, location);
        storer.write(object);
    }
    
    public static <T extends IOObject> T readFromRepository(String location, Class<T> clazz) throws OperatorCreationException, OperatorException {
        RepositorySource reader = OperatorService.createOperator(RepositorySource.class);
        reader.setParameter(RepositorySource.PARAMETER_REPOSITORY_ENTRY, location);
        return clazz.cast(reader.read());
    }
    
    public static String readFileAsString(String filePath) throws java.io.IOException{
        byte[] buffer = new byte[(int) new File(filePath).length()];
        FileInputStream f = new FileInputStream(filePath);
        f.read(buffer);
        return new String(buffer);
    }

    public static ExampleSet mergeExampleSets(ExampleSet a, ExampleSet b) throws OperatorCreationException, OperatorException {
        if(a == null) {
            return (ExampleSet)b.clone();
        }
        if(b == null) {
            return (ExampleSet)a.clone();
        }
        ExampleSetMerge merge = OperatorService.createOperator(ExampleSetMerge.class);
        List<ExampleSet> metaFeaturesCollection = new ArrayList<ExampleSet>();
        metaFeaturesCollection.add(a);
        metaFeaturesCollection.add(b);
        return merge.merge(metaFeaturesCollection);
    }

}
