/*
 *  RapidMiner PaREn Automatic System Construction Extension 
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.dfki.madm.mlwizard.functionality;

import java.util.Vector;

import com.rapidminer.example.ExampleSet;
import com.rapidminer.operator.OperatorCreationException;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.features.selection.AttributeWeightSelection;
import com.rapidminer.tools.OperatorService;
import com.rapidminer.tools.container.Pair;

import de.dfki.madm.mlwizard.Classifier;
import de.dfki.madm.mlwizard.KnowledgeBase;
import de.dfki.madm.mlwizard.metafeatures.MetaFeaturesOperator;

public class Regressioner {

	/**
	 * @brief result of the accuracy prediction step
	 * @author remat
	 */
	public static class RegressionResult {
	    /** meta-features of the new dataset */
	    public ExampleSet metaFeatures;
	    /** accuracy predictions for each classifier */ 
	    public Vector<Pair<Classifier, Double>> predictions;
	    
	    public RegressionResult(ExampleSet metaFeatures) {
	        this.metaFeatures = metaFeatures;
	        predictions = new Vector<Pair<Classifier,Double>>();
	    }
	}
	
	/**
	 * predict the accuracy for a given classifier and meta-features
	 * @param classifier the classifier which accuracy should be predicted
	 * @param metaFeatures the meta-features of the dataset which accuracy of the classifier should be predicted
	 * @return the predicted accuracy
	 * @throws OperatorCreationException
	 * @throws OperatorException
	 */
	public static double predict(Classifier classifier, ExampleSet metaFeatures) throws OperatorCreationException, OperatorException {
	    // -- preprocess meta-features (missing values, normalization, ...) --
	    metaFeatures = classifier.getPreprocessingModel().apply(metaFeatures);
	    // -- select used meta-features --
	    AttributeWeightSelection attributeSelection = OperatorService.createOperator(AttributeWeightSelection.class);
	    attributeSelection.getInputPorts().getPortByName("weights").receive(classifier.getMetaWeightsRegression());
	    metaFeatures = attributeSelection.apply(metaFeatures);
	    // -- apply regression model --
	    ExampleSet prediction = classifier.getRegressionModel().apply(metaFeatures);
	    return prediction.getExample(0).getNumericalValue(prediction.getAttributes().getPredictedLabel());
	}
	
	/**
	 * predicts the accuracies of all classifiers for one dataset
	 * @param knowledgeBase the knowledge base containing the classifiers
	 * @param dataset the dataset whose accuracy should be predicted
	 * @return all predicted accuracies
	 * @throws OperatorCreationException
	 * @throws OperatorException
	 */
	public static RegressionResult predict(KnowledgeBase knowledgeBase, ExampleSet dataset) throws OperatorCreationException, OperatorException {
	    // -- compute meta-features --
        MetaFeaturesOperator metaFeaturesOperator = OperatorService.createOperator(MetaFeaturesOperator.class);
        ExampleSet metaFeatures = metaFeaturesOperator.apply(dataset);
        RegressionResult result = new RegressionResult(metaFeatures);
        // -- predict every classifier --
	    for(Classifier classifier:knowledgeBase.getClassifiers()) {
	        result.predictions.add(new Pair<Classifier, Double>(classifier, predict(classifier, metaFeatures)));
	    }
	    return result;
	}
	
}
