package de.dfki.madm.mlwizard.functionality;

import java.io.IOException;
import java.util.HashMap;

import com.rapidminer.Process;
import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorCapability;
import com.rapidminer.operator.OperatorCreationException;
import com.rapidminer.operator.SimpleOperatorChain;
import com.rapidminer.operator.io.RepositorySource;
import com.rapidminer.operator.io.XrffExampleSource;
import com.rapidminer.operator.learner.AbstractLearner;
import com.rapidminer.operator.meta.ParameterSet;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.operator.preprocessing.filter.NominalToBinominal;
import com.rapidminer.operator.preprocessing.filter.NominalToNumeric;
import com.rapidminer.operator.preprocessing.normalization.Normalization;
import com.rapidminer.operator.preprocessing.normalization.RangeNormalizationMethod;
import com.rapidminer.tools.OperatorService;
import com.rapidminer.tools.XMLException;

import de.dfki.madm.mlwizard.Classifier;

/**
 * constructs the final classification system 
 * @author remat
 *
 */
public class SystemConstructor {
    
    /**
     * creates an operator and adds it to the process and connect its first input with the first output of the previous operator
     * @param process the process
     * @param clazz the class of the operator to be created
     * @param previsouOperator the previous operator
     * @return the newly created operator
     * @throws OperatorCreationException
     */
    public static <T extends Operator> T addOperator(Process process, Class<T> clazz, Operator previsouOperator) throws OperatorCreationException {
        T newOperator = OperatorService.createOperator(clazz);
        process.getRootOperator().getSubprocess(0).addOperator(newOperator);
        if(previsouOperator != null) {
            previsouOperator.getOutputPorts().getPortByIndex(0).connectTo(newOperator.getInputPorts().getPortByIndex(0));
        }
        return newOperator;
    }
    
    public static Process createProcess(Classifier classifier, ParameterSet parameterSet, String datasetLocation) throws OperatorCreationException, IOException, XMLException {
        return createProcess(classifier, parameterSet, datasetLocation, false);
    }


    /**
     * creates a complete process of creating a classification model for the given classifier, parameter set and dataset
     * @param classifier the classifier
     * @param parameterSet the parameter set that will be applied to the final process 
     * @param datasetLocation the location of the dataset
     * @return the created process
     * @throws OperatorCreationException
     * @throws XMLException 
     * @throws IOException 
     */
    public static Process createProcess(Classifier classifier, ParameterSet parameterSet, String datasetLocation, boolean xrff) throws OperatorCreationException, IOException, XMLException {
        Process process = new Process();

        Operator datasetReader;
        if(xrff) {
            datasetReader = addOperator(process, XrffExampleSource.class, null);
            datasetReader.setParameter(XrffExampleSource.PARAMETER_DATA_FILE, datasetLocation);
        } else {
            datasetReader = addOperator(process, RepositorySource.class, null);
            datasetReader.setParameter(RepositorySource.PARAMETER_REPOSITORY_ENTRY, datasetLocation);
        }

        if(!classifier.isComplex()) {
            Normalization normalization = addOperator(process, Normalization.class, datasetReader);
            normalization.setParameter(Normalization.PARAMETER_NORMALIZATION_METHOD, String.valueOf(Normalization.METHOD_RANGE_TRANSFORMATION));
            normalization.setParameter(RangeNormalizationMethod.PARAMETER_MIN, "0.0");
            normalization.setParameter(RangeNormalizationMethod.PARAMETER_MAX, "1.0");
            AbstractLearner learner = classifier.getLeanerInstance();
            boolean supportsNominal = learner.supportsCapability(OperatorCapability.BINOMINAL_ATTRIBUTES) && learner.supportsCapability(OperatorCapability.POLYNOMINAL_ATTRIBUTES);
            if(!supportsNominal) {
                NominalToBinominal nominalToBinominal = addOperator(process, NominalToBinominal.class, normalization);
                NominalToNumeric nominalToNumeric = addOperator(process, NominalToNumeric.class, nominalToBinominal);
                learner = addOperator(process, classifier.getLearner(), nominalToNumeric);
            } else {
                learner = addOperator(process, classifier.getLearner(), normalization);
            }
            learner.getOutputPorts().getPortByIndex(0).connectTo(process.getRootOperator().getSubprocess(0).getInnerSinks().getPortByIndex(0));
        } else {
            SimpleOperatorChain preProcessChain = addOperator(process, SimpleOperatorChain.class, datasetReader);
            preProcessChain.rename("Pre-Processing");
            Process preProcess = new Process(classifier.getPreProcess());
            InputPort input = preProcess.getRootOperator().getSubprocess(0).getInnerSources().getPortByIndex(0).getDestination();
            OutputPort output = preProcess.getRootOperator().getSubprocess(0).getInnerSinks().getPortByIndex(0).getSource();
            preProcessChain.getSubprocess(0).stealOperatorsFrom(preProcess.getRootOperator().getSubprocess(0));
            preProcessChain.getSubprocess(0).getInnerSources().getPortByIndex(0).connectTo(input);
            output.connectTo(preProcessChain.getSubprocess(0).getInnerSinks().getPortByIndex(0));
            
            SimpleOperatorChain learnChain = addOperator(process, SimpleOperatorChain.class, preProcessChain);
            learnChain.rename("Learning");
            Process learnProcess = new Process(classifier.getLearnProcess());
            input = learnProcess.getRootOperator().getSubprocess(0).getInnerSources().getPortByIndex(0).getDestination();
            output = learnProcess.getRootOperator().getSubprocess(0).getInnerSinks().getPortByIndex(0).getSource();
            learnChain.getSubprocess(0).stealOperatorsFrom(learnProcess.getRootOperator().getSubprocess(0));
            learnChain.getSubprocess(0).getInnerSources().getPortByIndex(0).connectTo(input);
            output.connectTo(learnChain.getSubprocess(0).getInnerSinks().getPortByIndex(0));
            
            learnChain.getOutputPorts().getPortByIndex(0).connectTo(process.getRootOperator().getSubprocess(0).getInnerSinks().getPortByIndex(0));
        }
        if(parameterSet != null) {
            parameterSet.applyAll(process,  new HashMap<String, String>());
        }
        return process;
    }
}
