package de.dfki.madm.mlwizard.metafeatures;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.ExampleSetFactory;
import com.rapidminer.operator.AbstractExampleSetProcessing;
import com.rapidminer.operator.OperatorCreationException;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.learner.tree.DecisionTreeLearner;
import com.rapidminer.operator.learner.tree.Edge;
import com.rapidminer.operator.learner.tree.Tree;
import com.rapidminer.operator.learner.tree.TreeBuilder;
import com.rapidminer.operator.preprocessing.discretization.BinDiscretization;
import com.rapidminer.tools.OperatorService;
import com.rapidminer.tools.container.Pair;

import de.dfki.madm.mlwizard.landmarking.AverageNodeLm;
import de.dfki.madm.mlwizard.landmarking.DecisionNodeLm;
import de.dfki.madm.mlwizard.landmarking.KNNLearnerLm;
import de.dfki.madm.mlwizard.landmarking.LandmarkingParameters;
import de.dfki.madm.mlwizard.landmarking.NaiveBayesLm;
import de.dfki.madm.mlwizard.landmarking.WorstNodeLm;

/**
 * @brief rapidminer operator for calculating meta-features of a dataset
 */
public class MetaFeaturesOperator extends AbstractExampleSetProcessing {

    private Map<String, AttributeStats> attributeStats;
    
    public MetaFeaturesOperator(OperatorDescription description) {
        super(description);
    }

    @Override
    public ExampleSet apply(ExampleSet exampleSet) throws OperatorException {
        try {
            return createExampleSet(compute(exampleSet));
        } catch (OperatorCreationException e) {
            e.printStackTrace();
            throw new OperatorException(e.getMessage(), e.getCause());
        }
    }

    /**
     * computes the meta-features of a given dataset
     * @param exampleSet the dataset
     * @return the computed meta-features
     * @throws OperatorCreationException
     * @throws OperatorException
     */
    public MetaFeatures compute(ExampleSet exampleSet) throws OperatorCreationException, OperatorException {
        Entropy entropy = new Entropy();
        ConditionalEntropy conditionalEntropy = new ConditionalEntropy();
        MutuaInformation mutuaInformation = new MutuaInformation();
        Kurtosis kurtosis = new Kurtosis();
        Skewness skewness = new Skewness();
        
        ExampleSet nominalExampleSet = discretize(exampleSet);
        
        attributeStats = new HashMap<String, AttributeStats>();
        for(Attribute attribute:exampleSet.getAttributes()) {
            attributeStats.put(attribute.getName(), new AttributeStats(exampleSet, nominalExampleSet, attribute));
        }
        Attribute label = exampleSet.getAttributes().getLabel();
        attributeStats.put(label.getName(), new AttributeStats(exampleSet, nominalExampleSet, label));
        
        MetaFeatures metaFeatures = new MetaFeatures();
        
        metaFeatures.add("samples", getNumberOfSamples(exampleSet));
        metaFeatures.add("attributes", getNumberOfAttributes(exampleSet));
        metaFeatures.add("classes", getNumberOfClasses(exampleSet));
        metaFeatures.add("numerical", getNumberOfNumircalAttributes(exampleSet));
        metaFeatures.add("nominal", getNumberOfNominalAttributes(exampleSet));
        metaFeatures.add("numericalRate", getRatioOfNumircalAttributes(exampleSet));
        metaFeatures.add("nominalRate", getRatioOfNominalAttributes(exampleSet));
        metaFeatures.add("dimensionality", getDimensionality(exampleSet));
        metaFeatures.addAll(getStats(getSymbols(exampleSet), "symbols"));
        
        // -- information-theoretic --
        metaFeatures.addAll(getStats(getPerAttribute(nominalExampleSet, entropy), "entropy"));
        metaFeatures.addAll(getStats(getPerAttribute(nominalExampleSet, conditionalEntropy), "conditional_entropy"));
        metaFeatures.addAll(getStats(getPerAttribute(nominalExampleSet, mutuaInformation), "mutual_information"));
        metaFeatures.add("class_entropy", entropy.compute(nominalExampleSet, nominalExampleSet.getAttributes().getLabel()));
        metaFeatures.add("equivalent_number_of_attributes", metaFeatures.get("class_entropy") / metaFeatures.get("mean_mutual_information"));
        metaFeatures.add("noise_signal_ratio", metaFeatures.get("mean_entropy") - metaFeatures.get("mean_mutual_information") / metaFeatures.get("mean_mutual_information"));
        
        // -- statistical --
        metaFeatures.addAll(getStats(getPerNumericalAttribute(exampleSet, kurtosis), "kurtosis"));
        metaFeatures.addAll(getStats(getPerNumericalAttribute(exampleSet, skewness), "skewness"));
        
        // -- landmarking --
        Map<String, String> parameters = new HashMap<String, String>();
        // -- cross validation parameters --
        parameters.put(LandmarkingParameters.PARAMETER_CROSS_VALIDATION, Boolean.toString(true));
        parameters.put(LandmarkingParameters.PARAMETER_LEAVE_ONE_OUT, Boolean.toString(true));
        parameters.put(LandmarkingParameters.PARAMETER_NUMBER_OF_VALIDATIONS, "10");
        parameters.put(LandmarkingParameters.PARAMETER_SAMPLING_TYPE, "shuffeled");
        parameters.put(LandmarkingParameters.PARAMETER_AVERAGE_PERFORMANCES_ONLY, Boolean.toString(true));
        // -- normalization parameters --
        parameters.put(LandmarkingParameters.PARAMETER_NORMALIZE_EXAMPLESET, Boolean.toString(false));
        parameters.put(LandmarkingParameters.PARAMETER_NORMALIZATION_METHOD, String.valueOf(LandmarkingParameters.METHOD_RANGE_TRANSFORMATION));
        parameters.put(LandmarkingParameters.PARAMETER_MIN, "0.0");
        parameters.put(LandmarkingParameters.PARAMETER_MAX, "0.1");

        metaFeatures.add("naive_bayes", new NaiveBayesLm(exampleSet, parameters).getAccuracy());
        metaFeatures.add("knn", new KNNLearnerLm(exampleSet, parameters).getAccuracy());
        metaFeatures.add("best_node", new DecisionNodeLm(exampleSet, parameters).getAccuracy());
        metaFeatures.add("average_node", new AverageNodeLm(exampleSet, parameters).getAccuracy());
        metaFeatures.add("worst_node", new WorstNodeLm(exampleSet, parameters).getAccuracy());
        
        // -- model-based --
        metaFeatures.addAll(treeBasedMetaFeatures(exampleSet));
        
        return metaFeatures;
    }

    /**
     * creates an example set out of a MetaFeatures object
     * @param data the meta-features
     * @return the meta-features as example set
     */
    public static ExampleSet createExampleSet(MetaFeatures data) {
        double[][] values = new double[1][data.size()];
        int i = 0;
        for(Pair<String, Double> entry:data) {
            values[0][i++] = entry.getSecond();
        }
        ExampleSet exampleSet = ExampleSetFactory.createExampleSet(values);
        Iterator<Attribute> it =  exampleSet.getAttributes().iterator();
        for(Pair<String, Double> entry:data) {
            it.next().setName(entry.getFirst());
        }
        return exampleSet;
    }
    
    private static double getNumberOfSamples(ExampleSet exampleSet) {
        return exampleSet.size();
    }
    
    private static double getNumberOfAttributes(ExampleSet exampleSet) {
        return exampleSet.getAttributes().size();
    }
    
    private static double getNumberOfClasses(ExampleSet exampleSet) {
        return exampleSet.getAttributes().getLabel().getMapping().size();
    }
    
    private static double getNumberOfNominalAttributes(ExampleSet exampleSet) {
        int n = 0;
        for(Attribute attribute:exampleSet.getAttributes()) {
            if(attribute.isNominal()) {
                n++;
            }
        }
        return n;
    }
    
    private static double getNumberOfNumircalAttributes(ExampleSet exampleSet) {
        int n = 0;
        for(Attribute attribute:exampleSet.getAttributes()) {
            if(attribute.isNumerical()) {
                n++;
            }
        }
        return n;
    }
    
    private static double getRatioOfNominalAttributes(ExampleSet exampleSet) {
        return getNumberOfNominalAttributes(exampleSet) / getNumberOfAttributes(exampleSet);
    }
    
    private static double getRatioOfNumircalAttributes(ExampleSet exampleSet) {
        return getNumberOfNumircalAttributes(exampleSet) / getNumberOfAttributes(exampleSet);
    }
    
    private static double getDimensionality(ExampleSet exampleSet) {
        return getNumberOfSamples(exampleSet) / getNumberOfAttributes(exampleSet);
    }

    
    private static Vector<Double> getSymbols(ExampleSet exampleSet) {
        Vector<Double> n = new Vector<Double>();
        for(Attribute attribute:exampleSet.getAttributes()) {
            if(attribute.isNominal()) {
                n.add((double)attribute.getMapping().size());
            }
        }
        return n;
    }
       

    /**
     * we use a vector as basis (and no map) since we want to preserve the order of the meta-features (for a better inspection by a user)
     */
    public static class MetaFeatures extends Vector<Pair<String, Double>> {
        private static final long serialVersionUID = 1438916065332388242L;
        public void add(String name, Double value) {
            add(new Pair<String, Double>(name, value));
        }
        public Double get(String name) {
            for(Pair<String, Double> entry:this) {
                if(entry.getFirst().equals(name)) {
                    return entry.getSecond();
                }
            }
            return Double.NaN;
        }
    }
    
    /******************** -- utility methods -- *********************/

    
    private interface Feature {
        public double compute(ExampleSet exampleSet, Attribute attribute);
    }
    
    private Vector<Double> getPerAttribute(ExampleSet exampleSet, Feature feature) {
        Vector<Double> n = new Vector<Double>();
        for(Attribute attribute:exampleSet.getAttributes()) {
            n.add(feature.compute(exampleSet, attribute));
        }
        return n;
    }
        
    private Vector<Double> getPerNumericalAttribute(ExampleSet exampleSet, Feature feature) {
        Vector<Double> n = new Vector<Double>();
        for(Attribute attribute:exampleSet.getAttributes()) {
            if(attribute.isNumerical()) {
                n.add(feature.compute(exampleSet, attribute));
            }
        }
        return n;
    }
    
    private class Entropy implements Feature {
        public double compute(ExampleSet nominalExampleSet, Attribute attribute) {
            AttributeStats stats = attributeStats.get(attribute.getName());
            double result = 0.0;
            for(String value : stats.frequencies.keySet()) {
                double frequency = stats.frequencies.get(value);
                result -= frequency * (Math.log(frequency) / Math.log(2));
            }
            return result;
        }
    }
    
    private class ConditionalEntropy implements Feature {
        public double compute(ExampleSet nominalExampleSet, Attribute attribute) {
            AttributeStats stats = attributeStats.get(attribute.getName());
            double result = 0.0;
            for (Pair<String, String> pair : stats.conditionalFrequencies.keySet()) {
                double frequencySingle = stats.frequencies.get(pair.getFirst());
                double frequencyPair = stats.conditionalFrequencies.get(pair);
                result += frequencyPair * (Math.log(frequencySingle / frequencyPair) / Math.log(2));
            }
            return result;
        }
    }
    
    private class MutuaInformation implements Feature {
        public double compute(ExampleSet nominalExampleSet, Attribute attribute) {
            AttributeStats stats = attributeStats.get(attribute.getName());
            AttributeStats classStats = attributeStats.get(nominalExampleSet.getAttributes().getLabel().getName());
            double result = 0.0;
            for (Pair<String, String> pair : stats.conditionalFrequencies.keySet()) {
                double frequencySingle = stats.frequencies.get(pair.getFirst());
                double frequencyClass = classStats.frequencies.get(pair.getSecond());
                double frequencyPair = stats.conditionalFrequencies.get(pair);
                result += frequencyPair * (Math.log(frequencyPair/(frequencySingle * frequencyClass)) / Math.log(2));
            }
            return result;
        }
    }
    
    private class Kurtosis implements Feature {
        public double compute(ExampleSet nominalExampleSet, Attribute attribute) {
            return attributeStats.get(attribute.getName()).kurtosis;
        }
    }
    
    private class Skewness implements Feature {
        public double compute(ExampleSet nominalExampleSet, Attribute attribute) {
            return attributeStats.get(attribute.getName()).skewness;
        }
    }
    
    private static ExampleSet discretize(ExampleSet exampleSet) throws OperatorCreationException, OperatorException {
        int nBins =  (int)Math.round(Math.sqrt(exampleSet.size()));
        BinDiscretization binDiscretization  = OperatorService.createOperator(BinDiscretization.class);
        binDiscretization.setParameter(BinDiscretization.PARAMETER_NUMBER_OF_BINS, String.valueOf(nBins));
        return binDiscretization.doWork(exampleSet);
    }
    
    /**
     * calculates min, max, mean, and standard deviation of a collection of values
     * @param values the values
     * @param suffix the suffix for the meta-feature names (e.g. min_<suffix>, max_<suffix>)
     * @return
     */
    private static MetaFeatures getStats(Collection<Double> values, String suffix) {
        double min = Double.MAX_VALUE;
        double max = Double.MIN_NORMAL;
        double mean = 0.0;
        double dev = 0.0;
        if(values.isEmpty()) {
            min = 0.0;
            max = 0.0;
        } else {
            for(Double v:values) {
                if(v < min) {
                    min = v;
                }
                if(v > max) {
                    max = v;
                }
                mean += v;
            }
            mean /= (double)values.size();
            for(Double v:values) {
                dev += (mean -v) * (mean - v);
            }
            dev = Math.sqrt(dev / (double)values.size());
        }
        MetaFeatures result = new MetaFeatures();
        result.add("min_"+suffix, min);
        result.add("max_"+suffix, max);
        result.add("mean_"+suffix, mean);
        result.add("dev_"+suffix, dev);
        return result;
    }
    
    /**
     * @brief little helper class to get access to the actual tree
     */
    public static class DecisionTreeLearnerOwn extends DecisionTreeLearner {
        public DecisionTreeLearnerOwn(OperatorDescription description) {
            super(description);
        }
        public TreeBuilder getTreeBuilder(ExampleSet exampleSet) throws OperatorException {
            return super.getTreeBuilder(exampleSet);
        }
    }
    
    /**
     * calculates the model/tree-based meta-features
     * @param exampleSet
     * @return
     * @throws OperatorCreationException
     * @throws OperatorException
     */
    private static MetaFeatures treeBasedMetaFeatures(ExampleSet exampleSet) throws OperatorCreationException, OperatorException {
        exampleSet = (ExampleSet)exampleSet.clone();
        DecisionTreeLearnerOwn treeLearner = OperatorService.createOperator(DecisionTreeLearnerOwn.class);
        TreeBuilder builder = treeLearner.getTreeBuilder(exampleSet); 
        Tree tree = builder.learnTree(exampleSet);
        TreeProperties treeProperties = treeProperties(tree, exampleSet);
        MetaFeatures result = new MetaFeatures();
        result.add("n_nodes", treeProperties.nNodes);
        result.add("n_leaves", treeProperties.nLeaves);
        result.addAll(getStats(treeProperties.lengths, "branch"));
        result.addAll(getStats(treeProperties.levels.values(), "level"));
        result.addAll(getStats(treeProperties.attributes.values(), "attribute"));
        return result;
    }

    /**
     * @brief encapsulation of tree properties used while traversing the tree
     */
    private static class TreeProperties {
        private Vector<Double> lengths = new Vector<Double>();
        private HashMap<Integer, Double> levels = new HashMap<Integer, Double>();
        private HashMap<String, Double> attributes = new HashMap<String, Double>();
        private double nNodes = 0;
        private double nLeaves = 0;
    }
    
    private static TreeProperties treeProperties(Tree tree, ExampleSet exampleSet) {
        TreeProperties treeProperties = new TreeProperties();
        for(Attribute attribute:exampleSet.getAttributes()) {
            treeProperties.attributes.put(attribute.getName(), 0.0);
        }
        traverse(tree, 0, treeProperties);
        return treeProperties;
    }
    
    /**
     * recursively traverses the tree and updates the tree properties
     * @param node the current node
     * @param depth the current depth
     * @param treeProperties the current tree properties
     */
    private static void traverse(Tree node, int depth, TreeProperties treeProperties) {
        depth++;
        treeProperties.nNodes++;
        if(!treeProperties.levels.containsKey(depth)) {
            treeProperties.levels.put(depth, 0.0);
        }
        treeProperties.levels.put(depth, treeProperties.levels.get(depth) + 1);
        if(node.isLeaf()) {
            treeProperties.nLeaves++;
            treeProperties.lengths.add((double)depth);
        } else {
            Iterator<Edge> children = node.childIterator();
            while(children.hasNext()) {
                Edge edge = children.next();
                String attribute = edge.getCondition().getAttributeName();
                treeProperties.attributes.put(attribute, treeProperties.attributes.get(attribute) + 1);
                traverse(edge.getChild(), depth, treeProperties);
            }
        }
    }

}

